
from os.path import join
import time

from epics import caget, caput, PV

from iexcode.instruments.utilities import dateandtime, print_warning_message
from iexcode.instruments.logfile import log_update
from iexcode.instruments.userCalcs import userStringSeq_pvs,userStringSeq_clear


def saveData_get_all(ioc_pv):
    """
    returns saveData info: 
    """
    saveData_info = {
        'fileSystem':caget(ioc_pv+"saveData_fileSystem",as_string=True),
        'subDir':caget(ioc_pv+"saveData_subDir",as_string=True),
        'scanNumber':caget(ioc_pv+"saveData_scanNumber"),
        'baseName':caget(ioc_pv+"saveData_baseName",as_string=True),
    }

    filepath = join(saveData_info['fileSystem'],saveData_info['subDir'])
    filepath=filepath.replace('//','/') 
    saveData_info['filepath'] = filepath

    return saveData_info
    

default_positioner_settling_time = 0.05
default_detector_settling_time = 0.1

class ScanRecord:
    """
    used for short hand scanning and to get info related to a scanRecord
    """

    def __init__(self,scan_ioc,**kwargs):
        """
        scan_ioc: full pv for the ioc, "29idARPES:", "29idc:"
        detector_dictionary: dictionary corresponding to detector num and pv to proc => {det_num:'det_pv'}
        trigger_dictionary: dictionary corresponding to trigger num and trigger pv => {trigger_num:'trigger_pv'}
        before_scan_pv: pv to be proc'ed be for the start of a scan
        after_scan_pv: pv to be proc'ed be for the start of a scan

        **kwargs
            detector_dictionary
            trigger_dictionary
            before_scan_pv
            after_scan_pv

        """


        self.ioc = scan_ioc

        if 'detector_dictionary' in kwargs:
            self.detector_dictionary = kwargs['detector_dictionary']
        if 'trigger_dictionary' in kwargs:
            self.trigger_dictionary = kwargs['trigger_dictionary']
        if 'before_scan_pv' in kwargs:
            self.before_scan_pv = kwargs['before_scan_pv']
        if 'after_scan_pv' in kwargs:
            self.after_scan_pv = kwargs['after_scan_pv']


        else:
            pass


        
        

##############################################################################################################
################################            MDA (ScanRecord) files and folders               ##############################
##############################################################################################################

    def fileNum(self): 
        """
        returns the current fileNum in the SaveData of ioc is scanning 
        or next if idle
        
        """
        
        return saveData_get_all(self.ioc)['scanNumber']

    def lastFileNum(self): 
        """
        returns the last saved fileNum in the SaveData of ioc

        previously: MDA_GetLastFileNum
        """

        fileNum = self.fileNum()
        return fileNum - 1


    def filepath(self):
        """
        returns the full filepath for the SaveData of ioc

        previously: MDA_CurrentDirectory
        """
        return saveData_get_all(self.ioc)['filepath']   

    def prefix(self):
        """
        returns the prefix for the SaveData of ioc

        previously: MDA_CurrentPrefix
        """
        return saveData_get_all(self.ioc)['baseName']   
   
    def scanRecord_run(self):
        """
        returns the current run by parsing the filepath in the SaveData for ioc

        previously: MDA_CurrentRun
        """
        filepath = self.filepath()
        m=filepath.find('data_29id')+len('data_29id')+2
        current_run=filepath[m:m+6]
        return current_run
    
    def scanRecord_user(self):
        """
        returns the current user by parsing the filepath in the SaveData for ioc

        previously: MDA_CurrentUser
        """
        subdir=saveData_get_all(self.ioc)['subDir']  
        m=subdir.find('/mda')
        if m == 0 : 
            current_user='Staff'
        elif m > 0: 
            current_user=subdir[1:m]
        else: 
            current_user=""
            print_warning_message("mda_user is empty string")
        return current_user   


##############################################################################################################
#############################          filling in the scan record               ##############################
##############################################################################################################


    ### before and after scans
    def _before_after_scan_set(self,**kwargs):
        """
        Clears the before_scan and after_scan in all dims
        Fills in the before_scan and after_after pv for kwargs['scan_dim'] only
        **kwargs
            scan_dim = 1 (default)
            before_scan_pv
            after_scan_pv
        """
        kwargs.setdefault('scan_dim',1)

        if 'before_scan_pv' in kwargs:
            self.before_scan_pv = kwargs['before_scan_pv']
        if 'after_scan_pv' in kwargs:
            self.before_scan_pv = kwargs['after_scan_pv']

        #Clearing all Before/Afters in all dims
        for dim in [1,2,3,4]:
            caput(self.ioc+"scan"+str(dim)+".BSPV","")
            caput(self.ioc+"scan"+str(dim)+".ASPV","")

        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])

        caput(scan_pv+".BSPV",self.before_scan_pv)
        caput(scan_pv+".BSCD",1)
        caput(scan_pv+".BSWAIT","Wait")

        caput(scan_pv+".ASPV",self.after_scan_pv)
        caput(scan_pv+".ASCD",1)
        caput(scan_pv+".ASWAIT","Wait")

    def default_after_scan_seq(self,**kwargs):
        """
        writes the user string sequence to happen at the end of a scan
        returns after_scan_pv = pv for userStringSeq for after scan

        uses links 1-4, link 10 is reserved for snake scans

        **kwargs
            seq_num: userStringSeq number in ioc => 10 (default)
            positioner_settling_time
            detector_settling_time

        Previously: AfterScan_StrSeq
        """
        kwargs.setdefault('seq_num',10)
        kwargs.setdefault('positioner_settling_time',default_positioner_settling_time)
        kwargs.setdefault('detector_settling_time',default_detector_settling_time)
        
        scan_dim = 1
        scan_pv = self.ioc+"scan"+str(scan_dim)

        seq_num=kwargs['seq_num']
        after_scan_pv = userStringSeq_pvs(self.ioc, seq_num)[0]
        self.after_scan_pv = after_scan_pv

        #clear and write the after scan user sequence
        userStringSeq_clear(self.ioc,seq_num)
        caput(after_scan_pv+".DESC","After Scan")

        #link 1 - clears positioners
        caput(after_scan_pv+".LNK1",scan_pv+'.P1AR'+" PP NMS")
        caput(after_scan_pv+".STR1","Clear pos&rdbk PV's, etc")

        #link 2 - sets positioner mode to to abosolute 
        caput(after_scan_pv+".LNK2",scan_pv+'.CMND'+" PP NMS")
        caput(after_scan_pv+".STR2","ABSOLUTE")

        #link 3 - sets positioner delay time to default 
        settling_time = kwargs['positioner_settling_time']
        caput(after_scan_pv+".LNK3",scan_pv+'.PDLY'+" PP NMS")
        caput(after_scan_pv+".DO3",settling_time)

        #link 4 - sets detector delay time to default 
        settling_time = kwargs['detector_settling_time']
        caput(after_scan_pv+".LNK4",scan_pv+'.DDLY'+" PP NMS")
        caput(after_scan_pv+".DO4",settling_time)

        return after_scan_pv+"+PROC"
    
    ### Resetting:
    def reset_scan_dim(self,scan_dim,**kwargs):
        """
        Resets all the scanRecords (scanDIM=1,2,3,4) for a given IOC 
            uses Reset_Scan()
        **kwargs
            detector_dictionary
            trigger_dictionary
            before_scan_pv
            after_scan_pv

        Previously: Reset_ScanAll
        """
        self.detectors_clear(scan_dim)
        self.triggers_clear(scan_dim)
        self.positioners_clear(scan_dim)

        if 'detector_dictionary' in kwargs:
            self.detectors_set(scan_dim,kwargs['detector_dictionary'])
        
        if 'trigger_dictionary' in kwargs:
            self.triggers_set(scan_dim,kwargs['trigger_dictionary'])

        self._before_after_scan_set(self,kwargs.update({'scan_dim':scan_dim}))

        self.detector_settling_time(scan_dim)
        self.positioner_settling_time(scan_dim)

    def reset_all(self):
        """
        Resets all the scanRecords (scanDIM=1,2,3,4) for a given IOC 
            uses Reset_Scan()
        **kwargs
            scaler='y', only for Kappa ioc

        Previously: Reset_ScanAll
        """

        for scan_dim in range(4,1):
            self.reset_scan_dim(scan_dim,{},{},'','')

        kwargs = {
            'detector_dictionary':self.detector_dictionary,
            'trigger_dictionary':self.trigger_dictionary,
            'before_scan_pv':self.before_scan_pv,
            'after_scan_pv':self.after_scan_pv
        }
        self.reset_scan_dim(1,**kwargs)

    ### default setting    
    def settings_defaults(self,scan_dim,verbose=True,**kwargs):
        """
        Reset scan settings to default: ABSOLUTE, STAY, 0.05/0.1 positioner/detector settling time
        
        Previous: Reset_Scan
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)

        self.positioner_absolute_mode(absolute=True,**kwargs)      # Absolute position
        self.positioner_after_scan(after="PRIOR POS",**kwargs)   # After scan: prior position  
        self.positioner_settling_time(**kwargs)# Positioner Settling time => default
        self.detector_settling_time(**kwargs)# Detector Settling time => default
        self.triggers_reset(verbose=False,**kwargs)

        if verbose:
            print("Scan record "+self.ioc+" reset to default settings.")

    ### positioners 
    def positioners_clear(self,scan_dim,verbose=True):
        """
        Clear all extra scan positioners

        Previously: Clear_Scan_Positioners
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        caput(scan_pv+".CMND",3)        # Clear all Positionners
        
        if verbose:
            print("\nAll extra positionners cleared")

    def positioner_settling_time(self,**kwargs):
        """
        sets the positioner settling time
        """
        kwargs.setdefault('scan_dim',1)
        kwargs.setdefault('positioner_settling_time',self.default_positioner_settling_time())

        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])

        caput(scan_pv+".PDLY",kwargs['positioner_settling_time'])     # Positioner Settling time
        return scan_pv+".PDLY",kwargs['positioner_settling_time']

    def positioner_after_scan(self,after="PRIOR POS",**kwargs):
        """
        sets the positioner after scan
        after = "PRIOR POS"
                "STAY"

        """
        kwargs.setdefault('scan_dim',1)
        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])
        caput(scan_pv+".PASM",after)

    def positioner_absolute_mode(self,absolute=True,**kwargs):
        """
        absolute = True
                 = False => relative 

        """
        kwargs.setdefault('scan_dim',1)

        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])
        if absolute:
            ab_state = 0
        else:
            ab_state = 1

        caput(scan_pv+".P1AR",ab_state)        # Absolute position

    def fillin(self,val_pv,rbv_pv,start,stop,steps_points,**kwargs):
        """
        Fills in the scanRecord for the positionner
            dim: scan dim
            val_pv: drive pv 
            rbv_pv: readback pv
            start: first point
            stop: last point
            step: step size (only used for positioner_num = 1, otherwise ignored)
        kwargs:
            num_points: False (default) => steps_points is the step size
                        True => steps_points is the number of points
            positioner_num: default => 1
            positioner_settling_time: default => 0.1
            detector_settling_time: default => 0.1

        Previously: Scan_FillIn
        """
        kwargs.setdefault('scan_dim',1)
        kwargs.setdefault("positioner_num",1)
        kwargs.setdefault('detector_settling_time',default_detector_settling_time)
        kwargs.setdefault('positioner_settling_time',default_positioner_settling_time)

        scan_dim = kwargs['scan_dim']
        self.progress(scan_dim)
    
        scan_pv = self.ioc+"scan"+str(scan_dim)
        posNum = kwargs['positioner_num']

        caput(scan_pv+".P"+str(posNum)+"PV",val_pv)
        caput(scan_pv+".R"+str(posNum)+"PV",rbv_pv)
        caput(scan_pv+".P"+str(posNum)+"SP",start*1.0)
        caput(scan_pv+".P"+str(posNum)+"EP",stop*1.0)
        if kwargs['positioner_num'] == 1:
            if kwargs['num_points'] :
                caput(scan_pv+".NPTS",steps_points)
            else:
                caput(scan_pv+".P1SI",steps_points*1.0)
        
        self.detector_settling_time(scan_dim,kwargs['detector_settling_time'])
        self.positioner_settling_time(scan_dim,kwargs['positioner_settling_time'])

        #checking that PVs and positioner limits are good
        self._check(scan_dim)


    ### detectors and detector settling time
    def detectors_clear(self,scan_dim,verbose=True):
        """
        Clear all scan detectors

        Previously: Clear_Scan_Detectors
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        for i in range(1,10):
            caput(scan_pv+".D0"+str(i)+"PV","")
        for i in range(10,71):
            caput(scan_pv+".D"+str(i)+"PV","")
        if verbose:
            print("\nAll detectors cleared")

    def detectors_set(self,scan_dim,detector_dictionary):
        """
        sets the detectors
        detector_dictionary = {detNum:pv}
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        
        if detector_dictionary != None:
            for det_num in detector_dictionary.keys():
                det_pv=caget(scan_pv+".D"+det_num+"PV")
                caput(det_pv,detector_dictionary[det_num])

    def detector_settling_time(self,**kwargs):
        """
        set the detector settling time
        """
        kwargs.setdefault('scan_dim',1)
        kwargs.setdefault('detector_settling_time',self.default_detector_settling_time())

        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])

        caput(scan_pv+".DDLY",kwargs['detector_settling_time'])
        return scan_pv+".DDLY",kwargs['detector_settling_time']

    ### triggers
    def triggers_set(self,scan_dim,trigger_dictionary):
        """
        Clear all scan detectors triggers
        
        trigger_dictionary = {trigger_num:pv}
        Previously: Clear_Scan_Triggers
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        if trigger_dictionary != None:
            if len(trigger_dictionary.keys())>0:
                for tigger_num in  trigger_dictionary.keys():
                        caput(scan_pv+".T"+str(tigger_num)+"PV",trigger_dictionary[tigger_num])  

    def triggers_clear(self,scan_dim,trigger_dictionary,verbose=True):
        """
        Clear all scan detectors triggers
        
        Previously: Clear_Scan_Triggers
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        inner_scan_trigger = self.ioc+"scan"+str(scan_dim-1)+".EXSC"

        if len(trigger_dictionary.keys())>0:   
            for dim in range(1,5):
                if dim==1 and scan_dim>1:
                    caput(scan_pv+".T1PV",inner_scan_trigger)
                else:
                    caput(scan_pv+'.T'+str(dim)+'PV',"")    

    def triggers_reset(self,verbose=True,**kwargs):
        """
        sets the  trigger1 to be the lower dim execute, does not change the other triggers
        does nothing if scan_dim = 1
        """
        kwargs.setdefault('scan_dim',1)
        scan_pv = self.ioc+"scan"+str(kwargs['scan_dim'])

        inner_scan_trigger = self.ioc+"scan"+str(kwargs['scan_dim']-1)+".EXSC"
        if kwargs['scan_dim'] > 1:
            caput(scan_pv+".T1PV",inner_scan_trigger)  #Resets the higher dimensional scans to trigger the inner scan

        if verbose:
            print("Scan record "+self.ioc+"Trigger 1 is reset")


    ### checks
    def check(self,scan_dim):
        """
        checks pvs connects and that positioners are within limits
        """
        self._check_pvs(scan_dim)
        if not self._check_limits(scan_dim):
            print("check positioner limits")

    def _check_pvs(self,scan_dim):
        """
        Check if any of the detectors or positions pvs are not connected

        Previously: Scan_Check
        """
        print('Checking if all detectors & positioners are connected...')
        scan_pv = self.ioc+"scan"+str(scan_dim)

        #Detectors
        for i in range(1,71):
            det_num=(str(i).zfill(2))
            det_pv=caget(scan_pv+".D"+det_num+"PV")
            if det_pv !='': 
                det=PV(det_pv); time.sleep(0.1)#smallest sleep to allow for PV traffic
                if not det.connected:
                    print("Detector "+det_num+" has a bad PV:  "+det.pvname+" not connected")
        #Positioners
        for i in range(1,5):
            pos_num=str(i)
            pos_pv=caget(scan_pv+".P"+pos_num+"PV")
            if pos_pv != '': 
                pos=PV(pos_pv)
                if not pos.connected:
                    print("Positioner "+pos_num+" has a BAD PV:  "+pos.pvname+" not connected")

    def _check_limits(self,scan_dim):
        """
        checks if the scan parameters are within the limits
        returns limits_ok (True / False)
        """
        scan_pv = self.ioc+"scan"+str(scan_dim)
        caput(scan_pv+'.CMND',1)
        time.sleep(.5)
        SMSG = caget(scan_pv+'SMSG',as_string=True)
        if len(SMSG) > 0:
            print_warning_message('Positioner out of limits \n     '+SMSG)
            limits_ok = False
        else:
            limits_ok = True
    

    ##############################################################################################################
    #############################          table scans               ##############################
    ##############################################################################################################  
    def fillin_table(self,scan_dim,val_pv,rbv_pv,myarray,**kwargs):
        """
        Fills in the scan record for table scans given positioner=posNum
        myarray can be generated by myarray=Scan_MakeTable(StartStopStepLists)

        kwargs:
            positioner_num: default => 1
            detector_settling_time: 
            positioner_settling_time

        Previously: Scan_FillIn_Table
        """
        kwargs.setdefault("positioner_num",1)
        kwargs.setdefault('detector_settling_time',default_detector_settling_time)
        kwargs.setdefault('positioner_settling_time',default_positioner_settling_time)

        self.progress(scan_dim)

        scan_pv = self.ioc+"scan"+str(scan_dim)
        posNum = kwargs['positioner_num']

        caput(scan_pv+".P"+str(posNum)+"SM","TABLE") 
        caput(scan_pv+".P"+str(posNum)+"PV",val_pv)         #Drive
        caput(scan_pv+".R"+str(posNum)+"PV",rbv_pv)         #Read  
        caput(scan_pv+".P"+str(posNum)+"PA",myarray)
        caput(scan_pv+'.NPTS',len(myarray))    

        self.detector_settling_time(scan_dim,kwargs['detector_settling_time']) 
        self.positioner_settling_time(scan_dim,kwargs['positioner_settling_time'])
        
        #checking that PVs and positioner limits are good
        self.check(scan_dim)
  
    def table_reset_after(self,scan_dim):
        """
        resets positioner settling time 0.1
        sets all positionitonser to linear
        clears positioners

        Previously: Scan_Reset_AfterTable
        """
        #Setting everything back
        scan_pv = self.ioc+"scan"+str(scan_dim)

        self.detector_settling_time(self.ioc,scan_dim) 

        for i in range(1,5):
            caput(scan_pv+".P"+str(i)+"SM","LINEAR") 

        self.positioners_clear(self.ioc,scan_dim)

    ##############################################################################################################
    #############################         progress, go and abort             ##############################
    ##############################################################################################################
    def progress(self,scan_dim,verbose=True):
        """
        Checks if a scan is in progress, and sleeps until it is done
        
        pv now includes the full 29idARPES:
        Previously: Scan_Progress
        """
        pv = self.ioc+"scan"+str(scan_dim)
        check=caget(pv+".FAZE")
        while (check!=0):
            if verbose:
                print(pv+" in progress - please wait...")
            time.sleep(30)
            check=caget(pv+".FAZE")

    def abort(self):
        """
        aborts a scan for the command line

        Previously: Scan_Abort
        """
        caput(self.ioc+"AbortScans.PROC",1)

    def go(self,verbose=True,**kwargs):
        """
        checks limits and then starts a scan for a given ioc and scan_dim
        
        Previously: part of Scan_Go
        """
        kwargs.setdefault('scan_dim',1)
        scan_dim = kwargs['scan_dim']

        self.progress(scan_dim)

        if verbose:
            for i in range(1,scan_dim+1):
                drive = caget(self.ioc+"scan"+str(i)+".P1PV")
                start = caget(self.ioc+"scan"+str(i)+".P1SP")
                stop  = caget(self.ioc+"scan"+str(i)+".P1EP")
                step  = caget(self.ioc+"scan"+str(i)+".P1SI")
                print('Scan'+str(i)+': '+drive+'= '+str(start)+' / '+str(stop)+' / '+str(step))
        
        if self._check_limits(scan_dim):
            filename = self.prefix(self.ioc)
            fileNum  = self.fileNum(self.ioc)
            print(filename+str(fileNum)+" started at ", dateandtime())
            scan_pv = self.ioc+"scan"+str(scan_dim)
            caput(scan_pv+".EXSC",1,wait=True,timeout=900000)  #pushes scan button
            log_update() # writes the log file
            print(filename+str(fileNum)+" finished at ", dateandtime())
            print('\n')
            

    ##############################################################################################################
    #############################          empty and time scans               ##############################
    ##############################################################################################################
 

    def empty_scan(self,**kwargs):
        """
        starts a scan with out a drive

        Previously: Scan_Empty_Go
        """
        kwargs.setdefault('execute',True)
        self.fillin("","",0,1,1,**kwargs)   
        if kwargs['execute']:
            self.go(**kwargs)
 
    def time_scan(self,duration_min,step_sec,**kwargs):
        """
        starts a scan with the readback as time in seconds

        **kwargs
            trigger_dictionary = {1:'29idb:ca5:read'} (default)
        
        Previously: Scan_Time_Go
        """

        stop=duration_min*60.0/step_sec
        self.fillin("","time",1,stop,1,**kwargs)

        print("Time scan - Settling time : "+str(step_sec))
        if kwargs['execute']:
            self.go(**kwargs)

    ##############################################################################################################
    #############################    2D scans typewriter/snake scans               ##############################
    ##############################################################################################################
    def fillin_2D(self,inner_loop_list,outer_loop_list,outer_scan_dim=2,**kwargs):
        """
        Fills in a the Scan Record for a 2D scan (Mesh),
        InnerMotorList=[val_pv1,rbv_pv1,start1,stop1,step1] #scanDIM=1
        OuterMotorList=[val_pv2,rbv_pv2,start2,stop2,step2] #scanDIM=2
        kwargs:
            num_points: False (default) => steps_points is the step size
                        True => steps_points is the number of points
            positioner_num: default => 1
            detector_settling_time: default => 0.1
            snake: coming soon            
        """
        kwargs.setdefault("snake",False)
        
        #fillin inner loop
        kwargs.update({'scan_dim':outer_scan_dim-1})
        self.fillin(inner_loop_list[0],inner_loop_list[1],inner_loop_list[2],inner_loop_list[3],inner_loop_list[4],**kwargs)
        #fillin outer loop
        kwargs.update({'scan_dim':outer_scan_dim})
        self.fillin(outer_loop_list[0],outer_loop_list[1],outer_loop_list[2],outer_loop_list[3],outer_loop_list[1],**kwargs)



    def snake_pv(self,snake_dim,enable=True):
        """
        used for snake scans (as opposed to typewriter scans)
        """
        snake_proc = self.snake_scan_userCalcOut(self.ioc,snake_dim=snake_dim,calc_num=1,enable=enable)
        return snake_proc


    def snake_scan_userCalcOut(self,snake_dim=1,calc_num=1,enable=False):
        """
        for snake scans you need to add this userCalcOut to the after scan
        it multiplies the step of the inner scan by -1 

        return the userCalcOut.PROC pv
        """
        snake_calc_pv = self.ioc+"userCalcOut"+str(calc_num) #Snake userCalc
        caput(self.ioc+"userCalcOutEnable.VAL",1) #global enable for userCalcs
        
        step_pv  = self.ioc+"scan"+str(snake_dim)+".P1SI"
        ### Setting up the user
        caput(snake_calc_pv+".DESC","Snake Calc")
        caput(snake_calc_pv+".INPA",step_pv+" NPP NMS")
        caput(snake_calc_pv+".CALC$","A*-1*B")
        caput(snake_calc_pv+".OUT",step_pv+" PP NMS")

        if enable:
            caput(snake_calc_pv+".B",1)
        else:
            caput(snake_calc_pv+".B",0)

        return snake_calc_pv+".PROC"